# Copyright 2023-2024 Advanced Micro Devices, Inc., and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from hatchet import GraphFrame
from hatchet.external.console import ConsoleRenderer

import pytest

perfetto_avail = True
try:
    # add noqa to line below because we need to disable these tests
    # for python < 3.6 because perfetto is not available. we do
    # not use this import locally because it is used inside
    # the PerfettoReader
    import perfetto  # noqa: F401
except ImportError:
    perfetto_avail = False


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_graphframe_python_source(omnitrace_python_source):
    """Validation test a GraphFrame object with known single rank, single thread data."""

    gf = GraphFrame.from_omnitrace(omnitrace_python_source)

    assert len(gf.dataframe) == 59

    gf = gf.squash()

    assert len(gf.dataframe) == 9


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_graphframe_mpi_aggregate(omnitrace_mpi_aggregate):
    """Validation test a GraphFrame object with multiple rank, multiple thread data in single file."""

    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_aggregate,
        include_category=["host", "pthread", "mpi"],
        exclude_category=["pthread", "mpi"],
        verbose=3,
        report=["all"],
    )

    assert len(gf.dataframe) == 32

    gf = gf.squash()

    assert len(gf.dataframe) == 8


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_graphframe_mpi_single(omnitrace_mpi_single):
    """Validation test a GraphFrame object with single rank, multiple thread data."""

    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_single, include_category=["host"], verbose=3, report=["none"]
    )

    assert len(gf.dataframe) == 562

    gf = gf.squash()

    assert len(gf.dataframe) == 52


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_graphframe_mpi_group(omnitrace_mpi_group):
    """Validation test a GraphFrame object with multiple rank, multiple thread data where data for each rank is in a separate file."""

    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_group, include_category=["host"], verbose=0, report=["all"]
    )

    assert len(gf.dataframe) == 6880

    gf = gf.squash()

    assert len(gf.dataframe) == 736


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_sampling_mpi_group(omnitrace_mpi_group):
    """Validate sampling data is preserved (each file has variable number of gotcha_wrap instances)"""

    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_group,
        include_category=["timer_sampling"],
        verbose=0,
        report=["all"],
    )

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time.inc",
        precision=3,
        name_column="name",
        expand_name=True,
        context_column="category",
        rank=0,
        thread=0,
        depth=8,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )

    output_v = "{}".format(output).split("\n")

    def get_line_count(key, data, ignore=[]):
        matching_lines = [x.strip() for x in data if key in x]
        for itr in ignore:
            matching_lines = [x for x in matching_lines if itr not in x]
        matching_count = len(matching_lines)

        print(
            "matching '{}' lines (n={}):\n    {}".format(
                key,
                matching_count,
                "\n    ".join([x for x in matching_lines if x is not None]),
            )
        )
        return matching_count

    assert get_line_count("_libc_start_main", output_v) == 472
    assert get_line_count("main", output_v, ["__libc_start_main", "run_main"]) == 479
    assert get_line_count("run_main", output_v) == 15
    assert get_line_count("PMPI_Init_thread", output_v) == 8
    assert get_line_count("mca_base_framework_open", output_v) == 8


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_tree_python_source(omnitrace_python_source):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_omnitrace(omnitrace_python_source)

    print(gf.tree("time.inc"))
    gf.calculate_exclusive_metrics()

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )

    print(output)


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_tree_mpi_aggregate(omnitrace_mpi_aggregate):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_aggregate, include_category=["host"], verbose=0, report=["none"]
    )
    gf = gf.squash()
    gf.calculate_exclusive_metrics()

    print(gf.tree("time.inc"))

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )

    print(output)


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_tree_mpi_group(omnitrace_mpi_group):
    """Sanity test a GraphFrame object with known data."""
    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_group, include_category=["host"], verbose=0, report=["none"]
    )

    print(gf.tree("time.inc"))
    gf.calculate_exclusive_metrics()

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )

    print(output)


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_read_attribute_mpi_group(omnitrace_mpi_group):
    """Generate a graphframe with only host data and then re-read but include mpi and pthread category data"""
    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_group,
        include_category=["host"],
    )

    assert gf.selected_categories() == ["host"]
    assert gf.available_categories() == [
        "host",
        "mpi",
        "overflow_sampling",
        "pthread",
        "timer_sampling",
    ]

    gf = gf.read(include_category=["host", "mpi", "pthread"]).squash()

    assert gf.selected_categories() == ["host", "mpi", "pthread"]

    assert len(gf.dataframe) == 1824


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_read_attribute_mpi_aggregate(omnitrace_mpi_aggregate):
    """Test the addition of "read" function attribute. ensure data is retained and works after garbage collection"""
    import gc

    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_aggregate, include_category=["host"], verbose=0, report=["none"]
    )

    assert gf.selected_categories() == ["host"]
    assert gf.available_categories() == [
        "host",
        "mpi",
        "overflow_sampling",
        "pthread",
        "timer_sampling",
    ]

    old_gf = gf.deepcopy()

    assert old_gf.selected_categories() == ["host"]
    assert old_gf.available_categories() == [
        "host",
        "mpi",
        "overflow_sampling",
        "pthread",
        "timer_sampling",
    ]

    gf = gf.read()
    gc.collect()  # make sure gf is still valid after garbage collection

    assert gf.selected_categories() == ["host"]
    assert gf.available_categories() == [
        "host",
        "mpi",
        "overflow_sampling",
        "pthread",
        "timer_sampling",
    ]
    assert len(gf.dataframe) == 32

    gf = gf.squash()
    gc.collect()
    gf.calculate_exclusive_metrics()

    assert len(gf.dataframe) == 8

    output = ConsoleRenderer(unicode=True, color=False).render(
        gf.graph.roots,
        gf.dataframe,
        metric_column="time",
        precision=3,
        name_column="name",
        expand_name=False,
        context_column="file",
        rank=0,
        thread=0,
        depth=10000,
        highlight_name=False,
        colormap="RdYlGn",
        invert_colormap=False,
    )

    print(output)


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_graphframe_to_literal(omnitrace_mpi_aggregate):
    """Test support for to_literal"""
    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_aggregate,
        exclude_category=["overflow_sampling", "timer_sampling"],
    ).squash()

    graph_literal = gf.to_literal()
    assert len(graph_literal) == len(gf.graph.roots)


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_default_metric_mpi_single_instrumentation(omnitrace_mpi_single):
    """Validation test for GraphFrame object using default metric field and to_{dot,flamegraph}"""
    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_single,
        exclude_category=["timer_sampling"],
    )

    for func in ["tree", "to_dot", "to_flamegraph"]:
        lhs = "{}\n{}".format(func, getattr(gf, func)(gf.default_metric))
        rhs = "{}\n{}".format(func, getattr(gf, func)())
        assert lhs == rhs


@pytest.mark.skipif(not perfetto_avail, reason="perfetto package not available")
def test_default_metric_mpi_single_sampling(omnitrace_mpi_single):
    """Validation test for GraphFrame object using default metric field and to_{dot,flamegraph}"""
    gf = GraphFrame.from_omnitrace(
        omnitrace_mpi_single,
        include_category=["timer_sampling"],
        max_depth=2,
    )

    for func in ["tree", "to_dot", "to_flamegraph"]:
        lhs = "{}\n{}".format(func, getattr(gf, func)(gf.default_metric))
        rhs = "{}\n{}".format(func, getattr(gf, func)())
        assert lhs == rhs



# Copyright 2022-2024 University of Oregon and other Hatchet Project
# Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pandas as pd

import hatchet.graphframe
from hatchet.node import Node
from hatchet.graph import Graph
from hatchet.frame import Frame
import glob
import json


class ApexReader:
    """Create a GraphFrame from a directory of APEX tasktree output JSON files.
    Example input:
    { "frame": {"name": "APEX MAIN", "type": "function", "rank": 0}, "metrics": {"time": 0.341264, "time (inc)": 0.341896, "min (inc)": 0.341896, "max (inc)": 0.341896, "sumsqr (inc)": 0.116893, "calls": 1.000000}, "children": [
     { "frame": {"name": "main", "type": "function", "rank": 0}, "metrics": {"time": 0.000059, "time (inc)": 0.000632, "min (inc)": 0.000632, "max (inc)": 0.000632, "sumsqr (inc)": 0.000000, "calls": 1.000000}, "children": [
      { "frame": {"name": "root_node", "type": "function", "rank": 0}, "metrics": {"time": 0.000217, "time (inc)": 0.000452, "min (inc)": 0.000452, "max (inc)": 0.000452, "sumsqr (inc)": 0.000000, "calls": 1.000000}, "children": [
       { "frame": {"name": "MPI_Recv", "type": "function", "rank": 0}, "metrics": {"time": 0.000027, "time (inc)": 0.000027, "min (inc)": 0.000002, "max (inc)": 0.000018, "sumsqr (inc)": 0.000000, "calls": 3.000000} },
       { "frame": {"name": "do_work", "type": "function", "rank": 0}, "metrics": {"time": 0.000111, "time (inc)": 0.000111, "min (inc)": 0.000111, "max (inc)": 0.000111, "sumsqr (inc)": 0.000000, "calls": 1.000000} },
       { "frame": {"name": "MPI_Send", "type": "function", "rank": 0}, "metrics": {"time": 0.000019, "time (inc)": 0.000019, "min (inc)": 0.000002, "max (inc)": 0.000008, "sumsqr (inc)": 0.000000, "calls": 6.000000} },
       { "frame": {"name": "get_next_work_item", "type": "function", "rank": 0}, "metrics": {"time": 0.000068, "time (inc)": 0.000079, "min (inc)": 0.000006, "max (inc)": 0.000031, "sumsqr (inc)": 0.000000, "calls": 8.000000}, "children": [
        { "frame": {"name": "get_work_items", "type": "function", "rank": 0}, "metrics": {"time": 0.000011, "time (inc)": 0.000011, "min (inc)": 0.000011, "max (inc)": 0.000011, "sumsqr (inc)": 0.000000, "calls": 1.000000} }
       ]
       }
      ]
      },
      { "frame": {"name": "MPI_Barrier", "type": "function", "rank": 0}, "metrics": {"time": 0.000120, "time (inc)": 0.000120, "min (inc)": 0.000017, "max (inc)": 0.000103, "sumsqr (inc)": 0.000000, "calls": 2.000000} }
     ]
     }
    ]
    }

    Return:
        (GraphFrame): graphframe containing data from dictionaries
    """

    def __init__(self, dirname):
        """Read from APEX tasktree json files in directory name

        dirname (string): Path to directory with tasktree.*.json files from APEX
        """
        self.warning = False
        self.dirname = dirname
        self.rank = 0
        self.ranks = []
        all_data = []
        files = sorted(glob.glob(dirname + "/tasktree.*.json"))
        for file in files:
            # get the rank number
            self.rank = file.split(".")[1]
            self.ranks.append(self.rank)
            # open text file in read mode
            with open(file) as json_file:
                data = json.load(json_file)
                all_data.append(data)
        self.graph_dict = all_data

    def parse_node_apex(
        self, frame_to_node_dict, node_dicts, child_dict, hparent, rank
    ):
        """Create node_dict for one node and then call the function
        recursively on all children.
        """

        if "rank" in child_dict["frame"]:
            rank = child_dict["frame"]["rank"]
            del child_dict["frame"]["rank"]
        if "type" not in child_dict["frame"]:
            child_dict["frame"]["type"] = "function"
        frame = Frame(child_dict["frame"])
        hnode = None
        found = False
        # See if the child exists already in the graph
        for child in hparent.children:
            if child.frame == frame:
                hnode = child
                found = True
                break
        # Doesn't exist, so add it
        if found is False:
            hnode = Node(frame, hparent, hnid=-1)

        # depending on the node type, the name may not be in the frame
        node_name = child_dict["frame"].get("name")
        if not node_name:
            node_name = child_dict["name"]
        node_dict = dict(
            {"node": hnode, "rank": rank, "name": node_name}, **child_dict["metrics"]
        )

        node_dicts.append(node_dict)
        frame_to_node_dict[frame] = hnode

        if not found:
            hparent.add_child(hnode)

        child_set = dict()
        if "children" in child_dict:
            for child in child_dict["children"]:
                # Validate the children names. We can potentially have duplicates when
                # we resolve OpenMP regions/loops that are outlined by compilers with
                # multiple loops. Each loop has a different address, but they resolve
                # to the same outlined function name, and with the same line number.
                # This can also happen if we resolve an address in a program or
                # library with no debug information, and there is no line number.
                # As a best effort, make the child name unique. It might not match
                # trees from multiple processes, though - the ordering might be different.
                c_name = child["frame"]["name"]
                if c_name in child_set.keys():
                    if not self.warning:
                        print(
                            rank,
                            ": WARNING! Duplicate child: ",
                            c_name,
                            "of parent",
                            child_dict["frame"]["name"],
                        )
                        self.warning = True
                    count = child_set[c_name] + 1
                    c_name = c_name + str(count)
                    child["frame"]["name"] = c_name
                    child_set[c_name] = count
                else:
                    child_set[c_name] = 1
                self.parse_node_apex(frame_to_node_dict, node_dicts, child, hnode, rank)

    def read(self):
        list_roots = []
        node_dicts = []
        frame_to_node_dict = {}
        frame = None
        top_node = None

        # start with creating a node_dict for each root
        for i in range(len(self.graph_dict)):
            rank = self.ranks[i]
            if "rank" in self.graph_dict[i]["frame"]:
                rank = self.graph_dict[i]["frame"]["rank"]
                del self.graph_dict[i]["frame"]["rank"]
            if "type" not in self.graph_dict[i]["frame"]:
                self.graph_dict[i]["frame"]["type"] = "function"
            frame = Frame(self.graph_dict[i]["frame"])
            graph_root = Node(frame, None, hnid=-1)
            if graph_root in list_roots:
                graph_root = top_node

            # depending on the node type, the name may not be in the frame
            node_name = self.graph_dict[i]["frame"].get("name")
            if not node_name:
                node_name = self.graph_dict[i]["name"]

            node_dict = dict(
                {"node": graph_root, "rank": rank, "name": node_name},
                **self.graph_dict[i]["metrics"],
            )
            node_dicts.append(node_dict)

            # Add the graph root to the list of roots, and save it
            # for the next tree to merge
            if graph_root not in list_roots:
                list_roots.append(graph_root)
                top_node = graph_root

            frame_to_node_dict[frame] = graph_root

            # call recursively on all children of root
            child_set = dict()
            if "children" in self.graph_dict[i]:
                for child in self.graph_dict[i]["children"]:
                    # Validate the children names. We can potentially have duplicates when
                    # we resolve OpenMP regions/loops that are outlined by compilers with
                    # multiple loops. Each loop has a different address, but they resolve
                    # to the same outlined function name, and with the same line number.
                    # This can also happen if we resolve an address in a program or
                    # library with no debug information, and there is no line number.
                    # As a best effort, make the child name unique. It might not match
                    # trees from multiple processes, though - the ordering might be different.
                    c_name = child["frame"]["name"]
                    if c_name in child_set.keys():
                        if not self.warning:
                            print(
                                rank,
                                ": WARNING! Duplicate child: ",
                                c_name,
                                "of parent",
                                node_name,
                            )
                            self.warning = True
                        count = child_set[c_name] + 1
                        c_name = c_name + str(count)
                        child["frame"]["name"] = c_name
                        child_set[c_name] = count
                    else:
                        child_set[c_name] = 1
                    self.parse_node_apex(
                        frame_to_node_dict, node_dicts, child, graph_root, rank
                    )

        graph = Graph(list_roots)
        graph.enumerate_traverse()

        exc_metrics = []
        inc_metrics = []
        for key in self.graph_dict[i]["metrics"].keys():
            if "(inc)" in key:
                inc_metrics.append(key)
            else:
                exc_metrics.append(key)

        dataframe = pd.DataFrame(data=node_dicts)

        # Make an index of all unique combinations of node, rank
        new_index = pd.MultiIndex.from_product(
            [dataframe["node"].unique(), dataframe["rank"].unique()],
            names=["node", "rank"],
        )
        dataframe.set_index(["node", "rank"], inplace=True)
        dataframe.sort_index(inplace=True)
        # For all missing values of [node, rank], add zero values for all columns
        dataframe = dataframe.reindex(new_index, fill_value=0)
        # Make our multi-index
        # Select rows where name is 0, we need to fix that.
        zeros = dataframe.loc[dataframe["name"] == 0]
        # For all columns with name '0', replace it with name from node.
        for row in zeros.itertuples(name="graphnodes"):
            name = row[0][0].frame["name"]
            index = row[0]
            dataframe.at[index, "name"] = name

        default_metric = "time (inc)"

        return hatchet.graphframe.GraphFrame(
            graph, dataframe, exc_metrics, inc_metrics, default_metric
        )



# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from functools import total_ordering

from .frame import Frame


def traversal_order(node):
    """Deterministic key function for sorting nodes in traversals."""
    return (node.frame, id(node))


@total_ordering
class Node:
    """A node in the graph. The node only stores its frame."""

    def __init__(self, frame_obj, parent=None, hnid=-1, depth=-1):
        self.frame = frame_obj
        self._depth = depth
        self._hatchet_nid = hnid

        self.parents = []
        if parent is not None:
            self.add_parent(parent)
        self.children = []

    def add_parent(self, node):
        """Adds a parent to this node's list of parents."""
        assert isinstance(node, Node)
        self.parents.append(node)

    def add_child(self, node):
        """Adds a child to this node's list of children."""
        assert isinstance(node, Node)
        self.children.append(node)

    def paths(self):
        """List of tuples, one for each path from this node to any root.

        Paths are tuples of node objects.
        """
        node_value = (self,)
        if not self.parents:
            return [node_value]
        else:
            paths = []
            for parent in self.parents:
                parent_paths = parent.paths()
                paths.extend([path + node_value for path in parent_paths])
            return paths

    def path(self, attrs=None):
        """Path to this node from root. Raises if there are multiple paths.

        This is useful for trees (where each node only has one path), as
        it just gets the only element from ``self.paths``.  This will
        fail with a MultiplePathError if there is more than one path to
        this node.
        """
        paths = self.paths()
        if len(paths) > 1:
            raise MultiplePathError("Node has more than one path: " % paths)
        return paths[0]

    def convert_path_to_str(self, path):
        """Takes a single path as a tuple of node
        objects and converts it to a tuple of strings.
        DO NOT USE THIS FUNCTION TO COMPARE NODES IN MOST CASES!
        """
        callpath_str = tuple()
        for node in path:
            callpath_str += (node.__str__(),)
        return callpath_str

    def dag_equal(self, other, vs=None, vo=None):
        """Check if DAG rooted at self has the same structure as that rooted at
        other.
        """
        if vs is None:
            vs = set()
        if vo is None:
            vo = set()

        vs.add(self._hatchet_nid)
        vo.add(other._hatchet_nid)

        # if number of children do not match, then nodes are not equal
        if len(self.children) != len(other.children):
            return False

        # sort children of each node by its frame
        ssorted = sorted(self.children, key=lambda x: x.frame)
        osorted = sorted(other.children, key=lambda x: x.frame)

        for self_child, other_child in zip(ssorted, osorted):
            # if frames do not match, then nodes are not equal
            if self_child.frame != other_child.frame:
                return False

            visited_s = self_child._hatchet_nid in vs
            visited_o = other_child._hatchet_nid in vo

            # check for duplicate nodes
            if visited_s != visited_o:
                return False

            # skip visited nodes
            if visited_s or visited_o:
                continue

            # recursive check for node equality
            if not self_child.dag_equal(other_child, vs, vo):
                return False

        return True

    def traverse(self, order="pre", attrs=None, visited=None):
        """Traverse the tree depth-first and yield each node.

        Arguments:
            order (str):  "pre" or "post" for preorder or postorder (default: pre)
            attrs (list or str, optional): if provided, extract these fields
                from nodes while traversing and yield them
            visited (dict, optional): dictionary in which each visited
                node's in-degree will be stored
        """
        if order not in ("pre", "post"):
            raise ValueError("order must be one of 'pre' or 'post'")

        if visited is None:
            visited = {}

        key = id(self)
        if key in visited:
            # count the number of times we reached
            visited[key] += 1
            return
        visited[key] = 1

        def value(node):
            return node if attrs is None else node.frame.values(attrs)

        if order == "pre":
            yield value(self)

        for child in sorted(self.children, key=traversal_order):
            for item in child.traverse(order=order, attrs=attrs, visited=visited):
                yield item

        if order == "post":
            yield value(self)

    def __hash__(self):
        return self._hatchet_nid

    def __eq__(self, other):
        return self._hatchet_nid == other._hatchet_nid

    def __lt__(self, other):
        return self._hatchet_nid < other._hatchet_nid

    def __gt__(self, other):
        return self._hatchet_nid > other._hatchet_nid

    def __str__(self):
        """Returns a string representation of the node."""
        return str(self.frame)

    def copy(self):
        """Copy this node without preserving parents or children."""
        return Node(frame_obj=self.frame.copy())

    @classmethod
    def from_lists(cls, lists):
        r"""Construct a hierarchy of nodes from recursive lists.

For example, this will construct a simple tree:

.. code-block:: python

    Node.from_lists(
        ["a",
            ["b", "d", "e"],
            ["c", "f", "g"],
        ]
    )

.. code-block:: console

         a
        / \
       b   c
     / |   | \
    d  e   f  g

And this will construct a simple diamond DAG:

.. code-block:: python

    d = Node(Frame(name="d"))
    Node.from_lists(
        ["a",
            ["b", d],
            ["c", d]
        ]
    )

.. code-block:: console

      a
     / \
    b   c
     \ /
      d

In the above examples, the 'a' represents a Node with its
`frame == Frame(name="a")`.
"""

        def _from_lists(lists, parent):
            if isinstance(lists, (tuple, list)):
                if isinstance(lists[0], Node):
                    node = lists[0]
                elif isinstance(lists[0], str):
                    node = Node(Frame(name=lists[0]))
                children = lists[1:]
                for val in children:
                    _ = _from_lists(val, node)
            elif isinstance(lists, str):
                node = Node(Frame(name=lists))
            elif isinstance(lists, Node):
                node = lists
            else:
                raise ValueError("Argument must be str, list, or Node: %s" % lists)

            if parent:
                node.add_parent(parent)
                parent.add_child(node)

            return node

        return _from_lists(lists, None)

    def __repr__(self):
        return "Node({%s})" % ", ".join(
            "%s: %s" % (repr(k), repr(v)) for k, v in sorted(self.frame.attrs.items())
        )


class MultiplePathError(Exception):
    """Raised when a node is asked for a single path but has multiple."""
